# -*- coding: cp1252 -*-

import      logging         as      log
from        htmlentitydefs  import  codepoint2name as entities
import      time
import      os
import      sys
import      ConfigParser

Ko              = 1024
CODAGE          = 'latin1'
DEFAULT_1WIRE   = 85.0
STRNAN          = 'nan'
NAN             = float(STRNAN)
CONFIG          = 'config'
PERIODE         = 'periode'
# Sections et tiquettes, comme dfini dans la config
COLLECTE        = 'collecte'
REMOTE          = 'remote'
TRACE           = 'trace'
AUTOMATE        = 'automate'
SERVEUR         = 'serveur'
PORT            = 'port'
DEBUG           = 'debug'
MACHINE         = 'machine'
READ_ADDR       = 'readAddr'

# Exception en cas d'erreur critique
class CriticalError(Exception): pass

def pluriel(val):
    return '' if val == 1 else 's'

# Pour affichage en seconde(s) ou minute(s) ou minute(s)+secondes
def outTime(val):
    if val < 60:
        return ('%d seconde' + pluriel(val)) % val
    minutes = val // 60
    secondes = val % 60
    if secondes == 0:
        return ('%d minute' + pluriel(minutes)) % minutes
    return ('%d minute' + pluriel(minutes) + ' et %d seconde' + pluriel(secondes)) % (minutes, secondes)

# Journalisation version module
def logVers(msg, version):
    ret = decodeUnicode(msg[0].upper() + msg[1:] + ', version du %s' % version)
    log.info(ret)
    return ret

# Lancement automatique du script
def runScript(main):
    try:
        main()
    except KeyboardInterrupt:
        # ferme la journalisation proprement
        log.critical(u'Arrt utilisateur par ^C')
    except CriticalError:
        # arrt volontaire sur erreur grave
        log.critical(u'Arrt sur erreur grave')
    finally:
        # ferme la journalisation
        log.shutdown()

# Lecture config
def readConfig(fileName, liste):
    cp = ConfigParser.RawConfigParser(liste)

    try:
        with open(fileName) as f:
            cp.readfp(f)
    except IOError:
        abort('Impossible d\'ouvrir le fichier de configuration %s' % fileName)

    return cp

# Arrt module
def abort(msg):
    '''Arrte le programme aprs avoir journalis le message
    msg : message  journaliser
    Retour : ne retourne pas
    '''
    log.critical(decodeUnicode(msg))
    raise CriticalError

# Dcodage suivant le set choisi
def decodeUnicode(msg):
    return msg.decode(CODAGE)

# Dcodage HTML, transforme une chaine de caractres "standard" en chaine HTML
def decodeHtml(msg):
    return ''.join('&%s;' % entities[ord(c)] if ord(c) in entities else c for c in msg)

# Initialisation du logging console et fichier
def initLog(journal):
    '''Configure la journalisation dans un fichier et sur la console.
    Par dfaut, journalise tout message au dessus de DEBUG.
    Les messages DEBUG sont journaliss si c'est demand dans la configuration.
    Retour : None
    '''  

    # si c'est dj configur (en cas de r-entrance), rien  faire
    if log.root.handlers:
        return
    
    # formatter des messages
    f = log.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s',
                      datefmt='%d/%m/%y %H:%M:%S')
    
    # handler console (stderr par dfaut)
    hc = log.StreamHandler()
    hc.setFormatter(f)
    hc.setLevel(log.DEBUG)

    # handler fichier journal
    hf = log.FileHandler(module(journal).extLog())
    hf.setFormatter(f)
    hf.setLevel(log.DEBUG)

    # logger par dfaut
    log.root.addHandler(hc)
    log.root.addHandler(hf)
    log.root.setLevel(log.INFO)

def logSansAff(journal):
    logfile = open(module(journal).extLog(), 'a', 0)
    sys.stdout = logfile
    sys.stderr = logfile

# endort un processus, secondes est la rcurrence et dlai un dphasage
def sleep(secondes, delai=0):
    t = time.localtime()
    time.sleep(secondes - ((t.tm_min * 60) + t.tm_sec) % secondes + delai)

# retourne le nom d'une appli sans son extension (ligne de commande)
def myselfSansExt():
    return module(os.path.basename(sys.argv[0])).sansExt()

class module(object):

    EXT_TXT         = '.txt'
    EXT_HTML        = '.html'
    EXT_PYTHON      = '.py'
    EXT_LOG         = '.log'
    EXT_CFG         = '.cfg'
    EXT_WATCHDOG    = '.txt'
    EXT_TMP         = '.tmp'

    def __init__(self, module):
        self.module = module

    # renvoie la chaine avant la dernire occurrence du '.' d'extension
    def sansExt(self):
        return os.path.splitext(self.module)[0]
        
    def extTxt(self):
        return self.module + self.EXT_TXT

    def extWatchDog(self):
        return self.module + self.EXT_WATCHDOG

    def extHtml(self):
        return self.module + self.EXT_HTML

    def extPy(self):
        return self.module + self.EXT_PYTHON

    def extLog(self):
        return self.module + self.EXT_LOG

    def extCfg(self):
        return self.module + self.EXT_CFG

    def extTest(self):
        return self.module + 'Test' + self.EXT_HTML

    def extDebug(self):
        return self.module + 'Debug' + self.EXT_HTML

    def extTmp(self):
        return self.module + self.EXT_TMP

    # criture des donnes dans le fichier
    def write(self, data):
        try:
            with open(self.module, 'w') as f:
                f.write(data)
                f.close()
        except:
            log.warning(u'Impossible d\'ouvrir le fichier : %s' % self.module)
            pass

# Gestion watchdog
class watchDog(object):

    FILE_WATCH      = 'Watch'
    FILE_DOG        = 'Dog'
    FILE_WATCHDOG   = FILE_WATCH + FILE_DOG
    TXT_WATCHDOG    = FILE_WATCHDOG

    def __init__(self, module):
        self.module = module

    # retourne le nom complet du fichier watchdog (avec son extension)
    def fileName(self):
        return module(self.module + self.FILE_WATCHDOG).extWatchDog()

    # Mise  jour du fichier watchdog
    def update(self):
        module(self.fileName()).write(self.TXT_WATCHDOG)

CFG_NAME = module(CONFIG).extCfg()
