# -*- coding: cp1252 -*-

''' Module de traitement des requtes venant du serveur et lies  l'automate
    La config contient:
        debug: bolen
        mot de passe (pwd)
        les caractristiques de l'automate (type, adresse des paramtres  lire et crire)
    Le fichier remote.html est un modle que Remote() modifie suivant les valeurs choisies, sortie par makeAndSendHtmlForm()
'''

import      logging         as      log
from        defGen          import  CFG_NAME, REMOTE, AUTOMATE, DEBUG, PORT, MACHINE, READ_ADDR, abort, readConfig, logVers, module
from        automates       import  M3Zelio                     # fichier contenant les automates grs
from        datetime        import  datetime, timedelta
from        collections     import  namedtuple
from        random          import  randrange
try:
    from    cStringIO       import StringIO
except ImportError:
    from    StringIO        import StringIO

VERSION             = '23/02/18'

# L'appli remote automate :-)
class Remote(object):

    # Messages
    PARAM_MSG       = 'Paramtres effectifs :'

    # Divers
    FALSE           = int(False)
    TRUE            = int(True)
    DEFVAL          = -1

    # Labels et units
    DT_ON           = 'DT ON'
    DT_OFF          = 'DT OFF'
    THRESHOLD       = 'Seuil'
    DURATION        = 'Dure'
    UNDISPLAYED     = '&nbsp;'
    LAYOUT          = ' '
    DEGREE          = LAYOUT + 'C'
    MINUTE          = LAYOUT + 'min' 

    # liste des donnes (invariable)
    '''
    TAG: tag champ, tansmise lors des changes
    VAL: valeur champ, modifie par _replace (copie objet Field)
    MIN, MAX:  valeurs min et max du champ
    STEP: incrment valeur champ
    DEFAULT: valeur par dfaut du champ
    TITLE: titre de la ligne affiche
    LABEL: label du champ (combo box)
    UNIT: unit affiche aprs le combo box
    CRLF: retour on non  la ligne (aprs x combo box)
    Format affichage ligne: TITLE + x fois (LABEL + combo box + UNIT)
    '''

    Field = namedtuple ('Field',
              'TAG              VAL     MIN MAX STEP    DEFAULT TITLE                          LABEL        UNIT            CRLF   ')
    FIELDS = (
        Field('dtOnNormal',     DEFVAL, 1,  20, 1,      10,     'Solaire mode normal',         DT_ON,       DEGREE,         FALSE   ),
        Field('dtOffNormal',    DEFVAL, 1,  10, 1,      3,      '',                            DT_OFF,      DEGREE,         TRUE    ),
        Field('dtOnDegrade',    DEFVAL, 1,  20, 1,      20,     'Solaire mode dgrad',        DT_ON,       DEGREE,         FALSE   ),
        Field('dtOffDegrade',   DEFVAL, 1,  10, 1,      10,     '',                            DT_OFF,      DEGREE,         FALSE   ),
        Field('seuilDegrade',   DEFVAL, 60, 95, 5,      80,     '',                            THRESHOLD,   DEGREE,         TRUE    ),
        Field('dtOnChauffage',  DEFVAL, 1,  20, 1,      5,      'Chauffage',                   DT_ON,       DEGREE,         FALSE   ),
        Field('dtOffChauffage', DEFVAL, 1,  10, 1,      1,      '',                            DT_OFF,      DEGREE,         TRUE    ),
        Field('seuilNocturne',  DEFVAL, 60, 95, 5,      80,     'Refroidissement nocturne',    THRESHOLD,   DEGREE,         FALSE   ),
        Field('dureeNoctune',   DEFVAL, 0,  60, 5,      30,     '',                            DURATION,    MINUTE,         TRUE    ),
        Field('tempMaxStock',   DEFVAL, 60, 95, 5,      95,     'Temprature maximale stock',  UNDISPLAYED, DEGREE,         TRUE    ),
        Field('vitCircu',       DEFVAL, 1,  3,  1,      1,      'Vitesse circulateur',         UNDISPLAYED, UNDISPLAYED,    TRUE    )
    )

    # label de champs reus dans la trame
    PWD_LABEL       = 'pwd'                                     # label pour identifier le mot de passe du client dans la trame
    TZ_LABEL        = 'tz'                                      # label pour identidier le timeZone du client dans la trame
    # mise en page du fichier HTML
    TITLE_WIDTH     = 200                                       # largeur titre
    LABEL_WIDTH     = 60                                        # largeur label (du combo)
    COMBO_WIDTH     = 100                                       # largeur (combo + unit)
    DEFVAL_WIDTH    = TITLE_WIDTH + LABEL_WIDTH                 # largeur champ bouton "Valeurs par defaut"
    DOC_WIDTH       = TITLE_WIDTH + 3 * (LABEL_WIDTH + COMBO_WIDTH) # largeur totale document

    Misc = namedtuple ('Misc',
               'PWD         TZ          TITLE_WIDTH     LABEL_WIDTH     COMBO_WIDTH     DEFVAL_WIDTH    DOC_WIDTH')
    MISC = Misc(PWD_LABEL,  TZ_LABEL,   TITLE_WIDTH,    LABEL_WIDTH,    COMBO_WIDTH,    DEFVAL_WIDTH,   DOC_WIDTH)

    def __init__(self):
        self.PARAM_COUNT    = len(self.FIELDS)                  # nombre de paramtres grs
        self.cfg            = self.config(self)                 # config de ce remote et de l'automate gr (statique)
        self.patternList    = self.getPatternList(self)         # lecture du formulaire modle, transform en liste (statique)

    # Traitement requte GET
    def get(self, File=None):
        PARAM_RD_ERR_MSG        = 'Erreur de lecture des paramtres de l\'automate, ' + self.PARAM_MSG
        # Lecture des valeurs courantes de l'automate
        if self.cfg.debug:
            data = randomize(self.FIELDS)                       # debug -> automate non gr, gnration de nombres alatoires
        else:
            data = self.cfg.controller.read(self.PARAM_COUNT)   # lecture automate

        if data is None:                                        # vrifie la lecture
            return self.makeAndSendHtmlForm(PARAM_RD_ERR_MSG)

        self.updateFields(self, data)                           # Mise  jour de la liste des valeurs courantes

        return self.makeAndSendHtmlForm(self.PARAM_MSG)         # Retourne le formulaire HTML modifi
    
    # Traitement requte POST
    def post(self, frame, File=None):
        ''' entre: frame, trame reue, dictionnaire {'ident':['valeur'],...}
        '''
        PWD_ERR_MSG         = 'Mot de passe erron, ' + self.PARAM_MSG
        FRAME_ERR_MSG       = 'Trame reue incorrecte, ' + self.PARAM_MSG
        PARAM_WR_ERR_MSG    = 'Erreur lors de l\'criture des paramtres de l\'automate, ' + self.PARAM_MSG

        # formats horodatage
        MM_TIME_FMT         = '%M'                              # minutes
        HHMM_TIME_FMT       = '%H' + MM_TIME_FMT                # heures + minutes

        # Vrifie la trame reue, prsence des champs 'pwd' et 'timeZone', prsence et valeur des champs de donnes, rcupre les donnes extraites de la trame
        List = [clientPwd, clientTimeZone, body] = self.isframeOk(frame, self.PWD_LABEL, self.TZ_LABEL, self.FIELDS)
        if List.count(None):
            return self.makeAndSendHtmlForm(FRAME_ERR_MSG)      # au moins une des valeurs est None (champ absent ou valeur hors limites)

        # Vrifie le mot de passe, paramtres d'appel = pwd serveur, pwd client, options (format horodatage, timeZone client, delta synchro horloges client/serveur)
        if not isPwdOk(self.cfg.pwd, clientPwd, HHMM_TIME_FMT, clientTimeZone, 2):
            return self.makeAndSendHtmlForm(PWD_ERR_MSG)        # Mot de passe erron

        # Constitue la trame  envoyer  l'automate
        data = []
        header = [int(2 ** self.PARAM_COUNT - 1)]               # entete trame = PARAM_COUNT bits  1 (commande multiplexeurs automate)
        data.extend(header)                                     # ajout entete trame
        data.extend(body)                                       # ajout corps trame (donnes)
        
        # crit les donnes dans l'automate et les verifie (en debug l'automate n'est pas gr)
        if not self.cfg.debug and self.cfg.controller.write(data) is None:
            return self.makeAndSendHtmlForm(PARAM_WR_ERR_MSG)   # erreur d'criture dans l'automate

        self.updateFields(self, data[len(header):])             # Mise  jour de la liste des valeurs courantes (sans l'entte)

        return self.makeAndSendHtmlForm(self.PARAM_MSG)         # Retourne le formulaire HTML modifi

    # Vrifie la trame reue, prsence des champs 'pwd' et 'timeZone', prsence et valeur des champs de donnes (List), retourne les donnes extraites ou une liste de x None
    def isframeOk(self, frame, pwdLabel, timeZoneLabel, List):
        try:
            clientPwd = frame[pwdLabel][0]                      # vrifie si le champ pwdLabel est dans la trame, clientPwd est une chaine
            clientTimeZone = int(frame[timeZoneLabel][0])       # vrifie si le champ timeZoneLabel est dans la trame, clientTimeZone est un entier, des minutes
        except:
            return 3 * [None]                                   # champ absent ou non conforme -> liste de 3 None en retour
        data = isDataFrameOk(frame, List)                       # vrifie prsence et valeur des champs de donnes, data est une liste d'entiers ou None en cas d'erreur
        return [clientPwd, clientTimeZone, data]                # liste des 3 valeurs

    # page HTML de test
    def test(self, File=None):
        return open(self.cfg.testFile)

    # page HTML de debug (nombres alatoires)
    def debug(self, File=None):
        self.updateFields(self, randomize(self.FIELDS))
        return self.makeAndSendHtmlForm('Debug :')

    # mme chose avec un fichier en sortie
    def debugInAFile(self, File=None):
        return module(self.cfg.debugFile).write(self.debug())

    # A partir de la liste modle gnre le formulaire HTML et l'envoie au client
    def makeAndSendHtmlForm(self, msg):
        '''
            Entre: msg, texte  afficher dans la barre de statut
            Sortie: La page HTML envoye
        '''

        List = list(self.patternList)                           # liste modle, modle conserv

        TAG_HEADER = '&'                                        # entte des tags

        # Dans la liste List remplace le champ TAG_HEADER+'FIELDS' par les valeurs dfinies dans Field et updateFields.fields
        tag, line, idx, pos = getTagLine(TAG_HEADER, 'FIELDS', List)
        List[idx] = ('\n'.join('%s = %d' %(fieldLabel, fieldIdx) for fieldIdx, fieldLabel in enumerate(self.Field._fields)) +
                     '\n' + tag + ' = \n[\n' + '],\n'.join('\t[' + ', '.join(('"%s"' if isinstance(fieldVal, str) else '%d')
                     % fieldVal for fieldVal in fieldLine) for fieldLine in self.updateFields.fields) + ']\n]\n')

        # Dans la liste List remplace le champ TAG_HEADER+'MISC' par les valeurs dfinies dans Misc et MISC
        tag, line, idx, pos = getTagLine(TAG_HEADER, 'MISC', List)
        List[idx] = ('\n'.join('%s = %d' % (miscLabel, miscIdx) for miscIdx, miscLabel in enumerate(self.Misc._fields)) +
                     '\n' + tag + ' = [' + ', '.join('"%s"' % miscVal for miscVal in self.MISC) + ']\n')

        # Dans la liste List remplace le champ TAG_HEADER+'MESSAGE' par le message  afficher dans la barre de tche
        tag, line, idx, pos = getTagLine(TAG_HEADER, 'MESSAGE', List)
        List[idx] = line[:pos] + msg + line[pos+len(TAG_HEADER+tag):]

        return outStream(List)                                  # traitement final

    # Mise  jour des valeurs courantes, updateFields.fields est une liste statique, copie modifie de FIELDS
    @staticmethod
    def updateFields(self, data):
        fields = []
        for idx, val in enumerate(data):
            fields.append(self.FIELDS[idx]._replace(VAL = int(val)))    # copie de la ligne de FIELDS aprs modification du champ VAL
        self.updateFields.fields = fields                       # liste statique

    # Lecture (unique, lors de la 1re instance) du formulaire HTML modle, le transforme en liste statique getPatternList.list
    @staticmethod
    def getPatternList(self):
        ''' Sortie: liste modifie
        '''

        myself = self.getPatternList

        if hasattr(myself, 'list'):
            return myself.list                                  # fichier dj lu
        
        myself.list = fileToListRead(self.cfg.sourceFile)       # lecture du formulaire modle, conversion en liste

        return myself.list

    # Lecture (unique, lors de la 1re instance) de la config, objet statique config.cfg en sortie
    @staticmethod
    def config(self, fileName=CFG_NAME):
        ''' Lit la configuration et la retourne dans un objet (statique) '''
        myself = self.config

        # lecture de la config si non dj lue
        if hasattr(myself, 'cfg'):                              # config dj lue
            return myself.cfg

        class Config(object):
            '''Objet pour contenir les attributs de configuration
            '''
            pass

        cfg = Config()

        # lecture de la config si non dj lue
            # lit le fichier de configuration
        cp = readConfig(fileName, {
            DEBUG:False,
            MACHINE: 'm3'
            })

        section = REMOTE
        # fichiers grs
        mod                             = module(section)
        cfg.sourceFile                  = mod.extHtml()
        cfg.testFile                    = mod.extTest()
        cfg.debugFile                   = mod.extDebug()

        # paramtres remote
        cfg.debug                       = cp.getboolean(section,    DEBUG)              # mode debug ou pas
        cfg.pwd                         = cp.get(section,           'pwd')              # mot de passe perso
        cfg.readAddr                    = cp.getint(section,        READ_ADDR)          # adresse des donnes automate  lire
        cfg.writeAddr                   = cp.getint(section,        'writeAddr')        # adresse des donnes automate  crire
        cfg.readCount = cfg.writeCount  = self.PARAM_COUNT                              # nombre de donnes gres
    
        # paramtres automates
        section = AUTOMATE
        cfg.machine                     = cp.get(section,           MACHINE)            # type d'automate
        cfg.port                        = cp.get(section,           PORT)               # port PC (USB) sur lequel l'automate est connect

        cfg.controller = Controllers(cfg.machine, cfg.port, cfg.readAddr, cfg.readCount, cfg.writeAddr, cfg.writeCount) # automate gr, M3 ou Zelio
        
        # Rappel de la config
        logVers(REMOTE, VERSION)
        log.info(u'     Configuration ' + REMOTE)
        log.info(u'         debug = %s' % cfg.debug)
        log.info(u'     Configuration ' + AUTOMATE)
        log.info(u'         %s = %s' % (AUTOMATE, cfg.machine))
        log.info(u'         port PC = %s' % cfg.port)
        log.info(u'         nombre de donnes  lire = %s' % cfg.readCount)
        log.info(u'         adresse 1re donne  lire = %s' % cfg.readAddr)
        log.info(u'         nombre de donnes  crire = %s' % cfg.writeCount)
        log.info(u'         adresse 1re donne  crire = %s' % cfg.writeAddr)

        myself.cfg = cfg                                        # objet statique en sortie
        return myself.cfg

################# Gestion des automates #################
    
class Controllers(object):
    def __init__(self, machine, port, readAddr, readCount, writeAddr, writeCount):
        self.controller = M3Zelio(machine, port, readAddr, readCount, writeAddr, writeCount)
                                  
    # Lecture des paramtres actifs de l'automate, retour = donnes lues ou None si erreur
    def read(self, count):
        self.controller.open()
        data = self.controller.read(count)
        self.controller.close()
        return data

    # Ecriture des paramtres actifs dans l'automate, retour = [] ou None si erreur
    def write(self, data):
        self.controller.open()
        ret = self.controller.write(data)
        self.controller.close()
        return ret

######################## fonctions diverses #########################

# Gnre des nombres altoires
def randomize(List):
    '''
        entre: liste de namedtuple contenant les champs MIN, MAX et STEP
        sortie: liste de donnes alatoires correspondant aux limites dfinies dans la liste fournie en entre
    '''
    data = []
    for line in List:
        data.append(randrange(line.MIN, line.MAX, line.STEP))   # suivant les valeurs extrmes et le step dfinis
    return data

# Sortie (streaming), liste en entre, fichier html en sortie
def outStream(List):
    form = ''.join(List)                                        # convertit la liste en chaine de caractres
    f = StringIO()                                              # sortie dans un fichier 'virtuel' (buffer)
    f.write(form)                                               # criture fichier
    f.seek(0)                                                   # dbut de fichier
    return form                                                 # pour autre utilisation

# Vrification du mot de passe client, ventuel horodatage avec calcul en UTC suivant le timeZone du client
def isPwdOk(serverPwd, clientPwd, timestampFmt='', clientTimeZone=0, clockDelta=0):
    ''' entres:
            serverPwd: chaine, pwd serveur (fichier de config ou autre)
            clientPwd: chaine (fourni par la trame)
            options:
                timestampFmt: chaine
                    '' -> pas d'horodatage, simple comparaison des pwd client et serveur
                    Autre -> pwd client horodat suivant le format spcifi, MM ou HHMM,
                             calculs effectus en UTC suivant le timeZone fourni par le client,
                             fuseaux horaires grs, partiellement si horodatage MM ou totalement si HHMM
                clientTimeZone: entier fourni par le client, +/- x minutes par rapport  l'heure UTC
                clockDelta: entier, permet de grer la diffrence de calage (quelques minutes) entre les horloges client et serveur
        sortie: boolen
    '''

    if timestampFmt == '':
        return clientPwd == serverPwd                           # pas d'horodatage

    now = datetime.utcnow()                                     # date + heure (UTC) du serveur, suite des calculs en UTC

    # Traitement du pwd client
    serverPwdLen = len(serverPwd)
    try:
        clientTimestamp = int(clientPwd[serverPwdLen:])         # horodatage en fin du pwd client, chaine de format MM ou HHMM convertie en entier
        # le pwd contient son horodatage, format de sortie xx...xMM ou xx...xHHMM (corrig UTC via le clientTimeZone)
        clientPwd = clientPwd[:serverPwdLen] + (datetime(now.year, now.month, now.day, clientTimestamp // 100, clientTimestamp % 100) + \
                                                timedelta(minutes = clientTimeZone)).strftime(timestampFmt)

        # Traitement du pwd serveur, format de sortie xx...xMM ou xx...xHHMM (UTC)
        delta = timedelta(minutes = clockDelta)                 # delta max entre horloges client et serveur
        before = serverPwd + (now - delta).strftime(timestampFmt)
        after = serverPwd + (now + delta).strftime(timestampFmt)

        return clientPwd <= after and clientPwd >= before       # en UTC, pwd client = pwd serveur avec horodatage (MM ou HHMM) dans la fourchette admise (clockDelta)
    except:
        return False                                            # horodatage client invalide, chaine vide ou non numriques

# Vrification de la prsence et de la valeur des champs de donnes dans la trame reue, chaine convertie en liste d'entiers
def isDataFrameOk(frame, List):
    ''' entres:
            frame: dictionnaire, {'ident':['valeur'],...}
            List: liste de namedtuple contenant les champs TAG, MIN et MAX
        sortie: liste d'entiers extraits de la trame si elle est correcte, None en cas d'erreur
    '''
    ret = []
    for line in List:                                           # verifie si tous les champs de donnes sont prsents dans la trame
        try:
            val = int(frame[line.TAG][0])
            if val < line.MIN or val > line.MAX:                # champ prsent et numrique, test de sa valeur
                return None                                     # valeur hors limites
            ret.append(val)                                     # ajout de la valeur extraite
        except:
            return None                                         # trame incorrecte, champ absent ou non numrique
    return ret                                                  # trame OK, retourne la liste des donnes extraites

# lit un fichier texte (html) ligne  ligne et le transforme en liste (sortie)
def fileToListRead(fileName):
    List = []
    try:
        with open(fileName, "r") as f:
            for line in f:
                List.append(line)
            f.close()
            return List
    except IOError:
        abort('Impossible d\'ouvrir le fichier %s' % fileName)

def getTagLine(tagHeader, tag, List):
    ''' Teste la prsence du champ tagHeader+tag dans la liste List, si trouv retourne le tag, la ligne dans cette liste,
        l'index de cette ligne dans cette liste ainsi que la position du champ dans la ligne de cette liste
        abort si tagHeader+tag absent
    '''
    for idx, line in enumerate(List):
        try:
            pos = line.index(tagHeader+tag)                     # recherche du champ
            return tag, line, idx, pos                          # champ trouv
        except:
            pass                                                # on passe  l'lment suivant de la liste

    abort("Le champ %s est introuvable dans le formulaire source" % (tagHeader+tag))    # erreur, tagHeader+tag absent
