# -*- coding: cp1252 -*-

''' Serveur minimaliste, ne gre les requtes que dans leur plus simple expression:
- GET, le fichier HTML modle est modifi suivant les paramtres actifs lus sur l'automate, puis envoy.
- POST, trame et mot de passe sont vrifis, page HTML et paramtres automate sont mis  jour si donnes OK.
- Connexion HTTP, non scurise, manque le 'S'...
'''

from        remote                  import  Remote
from        defGen                  import  CFG_NAME, REMOTE, SERVEUR, DEBUG, PORT, initLog, module, readConfig, runScript, logVers, myselfSansExt, logSansAff, Ko
import      sys
import      logging                 as      log
from        cgi                     import  parse_qs
from        wsgiref.simple_server   import  make_server
from        collections             import  namedtuple
import      os

VERSION = '31/01/18'

# Gre les demandes client
def serveur(environ, start_response):
    
    GET                     = 'GET'
    POST                    = 'POST'
    HEAD                    = 'HEAD'
    CONTENT                 = 'Content-Type'
    TEXT_HTML               = 'text/html'
    ICONE                   = 'image/x-icon'
    PNG                     = 'image/png'
    DEFAULT_HEADER          = [(CONTENT, TEXT_HTML)]

    # li au site
    FAVICON                 = 'favicon.ico'
    POWEREDBY               = 'python-powered.png'

    # li  l'appli remote
    REMOTE_TEST             = 'test'
    REMOTE_DEBUG            = 'debug'

    # status
    STATUS_PAGE_OK          = '200 OK'                      # page trouve
    STATUS_BAD_REQUEST      = '400 Bad Request'             # requte invalide
    STATUS_NOT_FOUND        = '404 Not Found'               # page introuvable

    # messages d'erreur
    MSG_BAD_REQUEST         = '<HTML>Erreur 400, requte invalide</HTML>'
    MSG_NOT_FOUND           = '<HTML>Erreur 404, page non trouve</HTML>'
    
#                    page web                       fichier d'entre            mthode POST    type contenu
#                                                               mthode GET
    desc_request = namedtuple ('desc_request',
                     'page                          fichier     get             post            type'       )
    DESC_REQUEST = (
        desc_request(page(''),                      None,       Remote().get,   Remote().post,  TEXT_HTML   ),
        desc_request(page(FAVICON),                 FAVICON,    getPicture,     None,           ICONE       ),
        desc_request(page(REMOTE),                  None,       Remote().get,   Remote().post,  TEXT_HTML   ),
        desc_request(page(REMOTE, FAVICON),         FAVICON,    getPicture,     None,           ICONE       ),
        desc_request(page(REMOTE, POWEREDBY),       POWEREDBY,  getPicture,     None,           PNG         ),
        desc_request(page(REMOTE, REMOTE_TEST),     None,       Remote().test,  None,           TEXT_HTML   ),
        desc_request(page(REMOTE, REMOTE_DEBUG),    None,       Remote().debug, None,           TEXT_HTML   ),
        )

    request = environ['REQUEST_METHOD']                     # la requte client
    path_info = environ['PATH_INFO']                        # page HTML demande     

    if request == HEAD:
        for line in DESC_REQUEST:
            if path_info == line.page:
                start_response(STATUS_PAGE_OK, DEFAULT_HEADER)
                return

        start_response(STATUS_NOT_FOUND, DEFAULT_HEADER)    # page introuvable
        return MSG_NOT_FOUND

    elif request == GET:
        for line in DESC_REQUEST:
            if path_info == line.page and line.get is not None:
                start_response(STATUS_PAGE_OK, [(CONTENT, line.type)])
                return line.get(line.fichier)               # appli spcifique

        start_response(STATUS_NOT_FOUND, DEFAULT_HEADER)    # page introuvable
        return MSG_NOT_FOUND
    
    elif request == POST:
        # criture des paramtres dans l'appli gre (automate)
        for line in DESC_REQUEST:
            if path_info == line.page and line.post is not None:
                start_response(STATUS_PAGE_OK, [(CONTENT, line.type)])
                size = int(environ.get('CONTENT_LENGTH', 0))
                body = environ['wsgi.input'].read(size)
                frame = parse_qs(body)                      # convertit body en dictionnaire (frame)
                return line.post(frame, line.fichier)       # appli spccifique

        start_response(STATUS_NOT_FOUND, DEFAULT_HEADER)    # page introuvable
        return MSG_NOT_FOUND

    start_response(STATUS_BAD_REQUEST, DEFAULT_HEADER)      # requte invalide
    return MSG_BAD_REQUEST

# Envoie des images (favicon et autre)
def getPicture(fileName):
    return open(fileName, 'rb')

# Gnre l'adresse de la page web (liste de longueur variable), format de sortie '/xxx/yyy/...'
def page(*args):
    return ''.join('/%s' % arg for arg in args)

def config(fileName=CFG_NAME):
    '''Lit la configuration et la retourne dans un objet
    '''
    class Config(object):
        '''Objet pour contenir les attributs de configuration
        '''
        pass

    # lit le fichier de configuration
    cp = readConfig(fileName, {
        'debug':False,
        })

    cfg = Config()

    section = SERVEUR
    cfg.debug   = cp.getboolean(section,    DEBUG)
    cfg.port    = cp.getint(section,        PORT)
    
    # met le niveau de journalisation en DEBUG si ncessaire
    if cfg.debug:
        log.root.setLevel(log.DEBUG)

    # Rappel de la config
    log.info(u'     Configuration ' + SERVEUR)
    log.info(u'         debug = %s' % cfg.debug)
    log.info(u'         port IP = %s' % cfg.port)

    return cfg

##### Point d'entre
def main():
    # nom de ce fichier sans son extension Python
    myself = myselfSansExt()

    # limitation de la taille du fichier log. En cas de dpassement on en conserve une partie (fin du log)
    logFileName = module(myself).extLog()                   # nom fichier log
    try:
        logSize = os.path.getsize(logFileName)              # taille log, exception si log inexistant (1er lancement)
        maxLogSize = 100 * Ko                               # taille max en Ko
        if logSize > maxLogSize:                            # taille log suprieure au max
            logFile = open(logFileName)
            logFile.seek(logSize - int(maxLogSize * 0.1))   # infos conserves en fin de log (0.1 du max)
            tmpFileName = module(myself).extTmp()           # nom fichier temporaire
            tmpFile = open(tmpFileName, "w")
            tmpFile.write(logFile.read())                   # criture de la fin du log dans le fichier temporaire
            tmpFile.close()
            logFile.close()
            os.remove(logFileName)
            os.rename(tmpFileName, logFileName)
    except:
        pass
    
    # configure la journalisation
    initLog(myself)

    # journalise la version
    logVers(SERVEUR, VERSION)

    # Lecture config
    cfg = config()

    if cfg.debug:
        Remote().debugInAFile()                             # gnre un fichier HTML de debug (donnes alatoires)
    else:
        logSansAff(myself)                                  # pour journaliser les requtes, sans affichage sur console serveur
    
    # Lancement serveur (boucle infinie en attente des requtes client)
    make_server('', cfg.port, serveur).serve_forever()

####################
# lancement automatique du script
if __name__ == '__main__':
    runScript(main)
