# -*- coding: cp1252 -*-
''' Processus de surveillance, ce module est un pre surveillant son fils...
Le pre (watch.py) lance le fils. Lorsque oprationnel, priodiquement le fils met  jour un fichier watchdog.
Le pre surveille l'heure de mise  jour de ce watchdog et si hors dlai tue son fils et le relance.
A chaque appli fils est associ un paramtre 'periode' dans la config, dfinissant la priodicit
du processus fils
'''


from            defGen          import  CFG_NAME, initLog, module, watchDog, readConfig, runScript, logVers, myselfSansExt, abort, outTime, PERIODE
import          subprocess
import          time
import          logging         as      log
import          os
import          sys

__all__ = ('main')

VERSION         = '30/01/18'
TOLERANCE       = 1.5

class Appli(object):
    '''Application gnrique  lancer et surveiller
    '''
    def __init__(self, pgm, watchFile, timeout):
        # sauve les donnes et dmarre le programme
        # et la surveillance
        self.pgm = pgm
        self.watchFile = watchFile
        self.timeout = timeout
        self.start()
        self.watch()
        
    def start(self):
        # dmarre (ou redmarre) le programme
        self.process = subprocess.Popen(self.pgm)
        log.debug(u'Appli demarre : %s' % self.pgm)

    def __del__(self):
        # kill the process if the object is deleted
        # (errors are ignored)
        self.process.terminate()

    def isActive(self):
        '''Teste si l'appli est toujours active
        (par dfaut teste juste s'il le programme s'excute)
        '''
        # teste juste si le programme s'excute toujours
        # (ce qui ne permet pas de dtecter les blocages)
        self.process.poll()
        return self.process.returncode is None

    def watch(self):
        '''Boucle en surveillant si le programme est actif
        Le relance s'il est arrt ou bloqu
        '''
        while 1:

            # attend jusqu'au contrle suivant
            time.sleep(self.timeout)

            # teste si le sous-process est actif
            if self.isActive():
                log.debug(u'Appli active : %s' % self.pgm)
            else:
                log.error(u'Appli inactive : %s' % self.pgm)

                # si le process est bloqu, on le tue
                self.process.poll()
                if not self.process.returncode:
                    log.error(u'Appli bloque,  tuer : %s' % self.pgm)
                    self.process.terminate()
                self.start()

#####################################################################

class generique(Appli):
    def isActive(self):
        '''Teste l'activit via la date de modification du fichier
        dcrivant l'installation
        '''
        # age en secondes de la dernire modif du fichier
        # on prend 50% de marge par rapport  la priode
        if os.path.exists(self.watchFile):
            age = time.time() - os.stat(self.watchFile).st_mtime
            return age <= self.timeout * TOLERANCE

        # fichier absent => inactivit par dfaut
        return False

def config(appliFils, section, fileName=CFG_NAME):
    '''Lit la configuration et la retourne dans un objet
    '''
    class Config(object):
        '''Objet pour contenir les attributs de configuration
        '''
        pass

    cfg = Config()

    # lit le fichier de configuration
    cp = readConfig(fileName, {})

    cfg.periode = cp.getint(section, PERIODE)

    # Rappel configuration
    log.info(u'     priodicit de l\'application surveille (%s) = %s ' % (appliFils, outTime(cfg.periode)))

    return cfg

# cre l'application
def main(appliFils, sectionConfig):
    # configure la journalisation de l'application pre (paramtre = nom de ce fichier sans son extension Python)
    initLog(myselfSansExt())

    # Indication de version
    logVers('Lanceur', VERSION)

    # Lecture du fichier de configuration (section dfinie dans le fichier de config)
    cfg = config(appliFils, sectionConfig)
    
    # lance l'appli gnrique (appli fils  surveiller, nom fichier watchdog, timeout)
    generique('python ' + appliFils, watchDog(module(appliFils).sansExt()).fileName(), cfg.periode)

####################
'''
28/11 : en faisant open/close  chaque fois, la collecte plante beaucoup plus.
Quand la collecte ne collecte plus,  0h00 le fichier de collecte est absent.
Le trac ne met plus  jour le fichier mais le parent ne semble pas le dtecter.
Au lieu de cela, a bloque la surveillance et le journal est vide  partir de l.
29/11 13:44
Gel du trac, mais bug dans parent.py qui empche le redmarrage, correction et relance
31/11 : problme  minuit car le graphe est renomm pour archivage, copie au lieu de renommage
02/12 : utilisation d'un fichier watchdog spcifique

MODIFS
- grouper la surveillance avec la trace
- utiliser la config
- utiliser le logging Python
'''
